'''Main file for entrypoint to package'''

import logging

import requests
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.capabilitystatement import CapabilityStatement
from fhir.resources.R4B.fhirtypes import Id
from fhir.resources.R4B.operationoutcome import OperationOutcome

from .helpers.capabilitystatement import get_supported_search_params, load_capability_statement
from .helpers.fhirfilter import filter_bundle
from .helpers.gapanalysis import run_gap_analysis
from .helpers.medicationhelper import expand_medication_references
from .helpers.documenthelper import expand_document_references
from .models.models import CustomFormatter, QuerySearchParams, SupportedSearchParams

logger: logging.Logger = logging.getLogger('fhirsearchhelper')
logger.setLevel(logging.INFO)
ch: logging.StreamHandler = logging.StreamHandler()
ch.setLevel(logging.INFO)
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)

# Epic IDs are sometimes more than the max_length of 64 set in the default
Id.configure_constraints(max_length=256)


def run_fhir_query(base_url: str = None, query_headers: dict[str, str] = None, search_params: QuerySearchParams = None, query: str = None, # type: ignore
                   capability_statement_file: str = None, capability_statement_url: str = None, debug: bool = False) -> Bundle | OperationOutcome | None: # type: ignore
    '''
    Entry function to run FHIR query using a CapabilityStatement and returning filtered resources
    WARNING: There is currently not a way to use a CapabilityStatement out of the box. See README.md of source for details.
    '''

    if debug:
        logger.info('Logging level is being set to DEBUG')
        logger.setLevel(logging.DEBUG)
        ch.setLevel(logging.DEBUG)
        logger.addHandler(ch)

    # Error handling
    if not base_url and not search_params and not query:
        raise ValueError('You must provide either a base_url and a dictionary of search parameters or the full query string in the form of <baseUrl>/<resourceType>?<param1>=<value1>&...')

    cap_state: CapabilityStatement = load_capability_statement(url=capability_statement_url, file_path=capability_statement_file)
    supported_search_params: list[SupportedSearchParams] = get_supported_search_params(cap_state)

    pretty_supported_search_params = {resource_params['resourceType']: [item['name'] for item in resource_params['searchParams']]
                                                                        for resource_params in [item.dict(exclude_none=True) for item in supported_search_params]}

    logger.debug(f'Supported search parameters for this server are: {pretty_supported_search_params}')

    if query:
        url_res, q_search_params = query.split('?')
        if url_res.split("/")[-1] not in pretty_supported_search_params:
            logger.error(f'Resource {url_res.split("/")[-1]} is not supported for searching, returning empty Bundle')
            return Bundle(**{'type': 'searchset', 'total': 0, 'link': [{'relation': 'self', 'url': url_res}]})
        if not q_search_params:
            logger.error('No search params, Epic does not support pulling all resources of a given type with no search parameters. Please refine your query.')
            new_query_response = requests.get(f'{url_res}', headers=query_headers)
            if new_query_response.status_code == 403:
                logger.error(f'The query responded with a status code of {new_query_response.status_code}')
                if 'WWW-Authenticate' in new_query_response.headers:
                    logger.error(f'WWW-Authenticate Error: {new_query_response.headers["WWW-Authenticate"]}')
                return None
            elif new_query_response.status_code == 400:
                return OperationOutcome(**new_query_response.json())
            else:
                return None

        base_url = '/'.join(url_res.split('/')[:-1])
        q_resource_type = url_res.split('/')[-1]
        search_params_list: list[str] = q_search_params.split('&')
        search_params_dict: dict[str, str] = {item.split('=')[0]: item.split('=')[1] for item in search_params_list}
        search_params: QuerySearchParams = QuerySearchParams(resourceType=q_resource_type, searchParams=search_params_dict)

    logger.debug(f'Search parameters for this request are: {search_params}')

    gap_output: list[str] = run_gap_analysis(supported_search_params=supported_search_params, query_search_params=search_params)

    logger.debug(f'Gap output from these two sets of search parameters is: {gap_output}')

    new_query_params_str = '&'.join([f'{key}={value}' for key, value in search_params.searchParams.items() if key not in gap_output])
    if new_query_params_str:
        new_query_string = f'{search_params.resourceType}?{new_query_params_str}'
    else:
        new_query_string = search_params.resourceType

    # new_query_string = new_query_string.replace('%7C', '|')

    logger.debug(f'New query string is {new_query_string}')

    logger.debug(f'Making request to {base_url}/{new_query_string}')
    new_query_response = requests.get(f'{base_url}/{new_query_string}', headers=query_headers)
    if new_query_response.status_code != 200:
        logger.error(f'The query responded with a status code of {new_query_response.status_code}')
        if 'WWW-Authenticate' in new_query_response.headers:
            logger.error(f'WWW-Authenticate Error: {new_query_response.headers["WWW-Authenticate"]}')
        return None

    new_query_response_bundle: Bundle | None = Bundle.parse_obj(new_query_response.json())

    if not new_query_response_bundle.entry:
        return new_query_response_bundle

    if 'operationoutcome' in [entry.resource.resource_type.lower() for entry in new_query_response_bundle.entry]:  #type: ignore
        new_query_response_bundle.entry = []

    if 'MedicationRequest' in new_query_string:
        logger.debug('Resources are of type MedicationRequest, proceeding to expand MedicationReferences')
        new_query_response_bundle = expand_medication_references(input_bundle=new_query_response_bundle, base_url=base_url, query_headers=query_headers)
        if not new_query_response_bundle:
            return None

    if 'DocumentReference' in new_query_string:
        logger.debug('Resources are of type DocumentReference, proceeding to expand DocumentReferences')
        new_query_response_bundle = expand_document_references(input_bundle=new_query_response_bundle, base_url=base_url, query_headers=query_headers)
        if not new_query_response_bundle:
            return None

    logger.debug(f'Size of bundle before filtering is {new_query_response_bundle.total} resources')
    filtered_bundle: Bundle = filter_bundle(input_bundle=new_query_response_bundle, search_params=search_params, gap_analysis_output=gap_output)
    logger.debug(f'Size of bundle after filtering is {filtered_bundle.total} resources')

    return filtered_bundle



'''File to handle all operations around OperationOutcomes'''

import logging

logger: logging.Logger = logging.getLogger('fhirsearchhelper.operationoutcomehelper')

def handle_operation_outcomes(resource: dict) -> None:
    issue_codes = [issue['code'] for issue in resource['issue'] if 'code' in issue]
    issue_details = [issue['details']['text'] for issue in resource['issue'] if 'details' in issue and 'text' in issue['details']]
    issue_diagnostics = [issue['diagnostics'] for issue in resource['issue'] if 'diagnostics' in issue]
    logger.warning(f'There was an OperationOutcome in the return bundle with codes {issue_codes}')
    logger.warning(f'There was an OperationOutcome in the return bundle with issue details of {issue_details}')
    logger.warning(f'There was an OperationOutcome in the return bundle with issue diagnostics of {issue_diagnostics}')


import pytest

from fhirsearchhelper.helpers.capabilitystatement import load_capability_statement, get_supported_search_params
from fhirsearchhelper.models.models import SupportedSearchParams
from fhir.resources.R4B.capabilitystatement import CapabilityStatement

def test_load_capability_statement_url() -> None:

    cs: CapabilityStatement = load_capability_statement(url='https://hapi.fhir.org/baseR4/metadata')

    assert cs.resource_type == 'CapabilityStatement'
    assert str(cs.implementation.url) == 'https://hapi.fhir.org/baseR4' #type: ignore


def test_load_capability_statement_file_path() -> None:

    cs: CapabilityStatement = load_capability_statement(file_path='epic_r4_metadata_edited.json')

    assert cs.resource_type == 'CapabilityStatement'
    assert str(cs.implementation.url) == 'https://fhir.epic.com/interconnect-fhir-oauth/api/FHIR/R4' #type: ignore


def test_load_capability_statement_no_args() -> None:

    with pytest.raises(ValueError):
        load_capability_statement()

def test_load_capability_statement_both_args() -> None:

    cs: CapabilityStatement = load_capability_statement(url='https://hapi.fhir.org/baseR4/metadata', file_path='epic_r4_metadata_edited.json')

    assert cs.resource_type == 'CapabilityStatement'
    assert str(cs.implementation.url) == 'https://hapi.fhir.org/baseR4' #type: ignore


def test_get_supported_search_params_capstate() -> None:

    cs: CapabilityStatement = load_capability_statement(file_path='epic_r4_metadata_edited.json')

    ssps: list[SupportedSearchParams] = get_supported_search_params(cs=cs)

    assert ssps
    assert isinstance(ssps, list)
    assert all([isinstance(sps, SupportedSearchParams) for sps in ssps])
    assert all([resource.type in [sps.resourceType for sps in ssps] for resource in cs.rest[0].resource if 'searchParam' in resource.dict(exclude_none=True)]) #type: ignore


'''File for custom models'''

from fhir.resources.R4B.capabilitystatement import CapabilityStatementRestResourceSearchParam
from pydantic import BaseModel
import logging


class CustomFormatter(logging.Formatter):
    grey: str = "\x1b[38;21m"
    green: str = "\x1b[32m"
    yellow: str = "\x1b[33m"
    red: str = "\x1b[31m"
    bold_red: str = "\x1b[31;1m"
    reset: str = "\x1b[0m"
    format: str = '{asctime}   {levelname:8s} --- {name}: {message}'

    FORMATS: dict[int, str] = {
        logging.DEBUG: grey + format + reset,
        logging.INFO: green + format + reset,
        logging.WARNING: yellow + format + reset,
        logging.ERROR: red + format + reset,
        logging.CRITICAL: bold_red + format + reset
    }

    def format(self, record) -> str:
        log_fmt: str = self.FORMATS.get(record.levelno)
        formatter: logging.Formatter = logging.Formatter(log_fmt, '%m/%d/%Y %I:%M:%S %p', style='{')
        return formatter.format(record)


class SupportedSearchParams(BaseModel):

    resourceType: str
    searchParams: list[CapabilityStatementRestResourceSearchParam]


class QuerySearchParams(BaseModel):

    resourceType: str
    searchParams: dict[str, str]



from fhirsearchhelper.helpers.fhirfilter import filter_bundle
from fhirsearchhelper.models.models import QuerySearchParams

from fhir.resources.R4B.bundle import Bundle


def test_filter_bundle_new() -> None:
    input_bundle: Bundle = Bundle.parse_file(
        "./tests/resources/FHIRFilterInputBundle.json"
    )
    search_params: QuerySearchParams = QuerySearchParams(
        resourceType="Condition",
        searchParams={
            "patient": "eZ5-7rYdWqgv3jSgIvx.SPw3",
            "category": "encounter-diagnosis",
            "code": "http://snomed.info/sct|110483000",
        },
    )
    gap_analysis_output: list[str] = ["code"]

    filtered_bundle: Bundle = filter_bundle(
        input_bundle=input_bundle,
        search_params=search_params,
        gap_analysis_output=gap_analysis_output,
    )

    assert filtered_bundle.total < input_bundle.total
    assert filtered_bundle.total == len(filtered_bundle.entry)

    for entry in filtered_bundle.entry:
        assert "eZ5-7rYdWqgv3jSgIvx.SPw3" == entry["resource"]["subject"]["reference"].split("/")[1]
        assert "encounter-diagnosis" in [cat["coding"][0]["code"] for cat in entry["resource"]["category"]]
        assert any(["http://snomed.info/sct|110483000".split("|")[1] == coding["code"] for coding in entry["resource"]["code"]["coding"]])



'''File to perform the search parameter gap analysis'''

import logging

from fhir.resources.R4B.capabilitystatement import CapabilityStatementRestResourceSearchParam

from ..models.models import QuerySearchParams, SupportedSearchParams

logger: logging.Logger = logging.getLogger('fhirsearchhelper.gapanalysis')


def run_gap_analysis(supported_search_params: list[SupportedSearchParams], query_search_params: QuerySearchParams) -> list[str]:
    resource_type = query_search_params.resourceType

    resouce_supported_search_params: SupportedSearchParams = list(filter(lambda x: x.resourceType == resource_type, supported_search_params))[0]

    query_params_names = list(query_search_params.searchParams.keys())
    temp_supported_params_names: list[str] = [rsc.name for rsc in resouce_supported_search_params.searchParams]

    supported_params_names = []
    for name in temp_supported_params_names:
        search_param_obj: CapabilityStatementRestResourceSearchParam = list(filter(lambda x: x.name == name, resouce_supported_search_params.searchParams))[0]
        if search_param_obj.extension:
            filtered_ext_list = list(filter(lambda x: x.url == 'true-when', search_param_obj.extension)) # type: ignore
            if filtered_ext_list:
                if '==' in filtered_ext_list[0].valueString:
                    field, value = filtered_ext_list[0].valueString.split('==')
                    if field in query_params_names and query_search_params.searchParams[field] == value:
                        supported_params_names.append(name)
                elif ' in ' in filtered_ext_list[0].valueString:
                    field, values = filtered_ext_list[0].valueString.split(' in ')
                    values = [value.strip() for value in  values.strip('[').strip(']').split(',')]
                    if (field in query_params_names
                        and len(query_search_params.searchParams[field].split(',')) > 1
                        and all([sp_value in values for sp_value in query_search_params.searchParams[field].split(',')])):
                        supported_params_names.append(name)
                        continue
                    if field in query_params_names and query_search_params.searchParams[field] in values:
                        supported_params_names.append(name)
        else:
            supported_params_names.append(name)

    params_not_supported = [name for name in query_params_names if name not in supported_params_names]

    return params_not_supported


'''File to handle all operations around Medication-related Resources'''

import logging
from copy import deepcopy

import requests
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType

logger: logging.Logger = logging.getLogger('fhirsearchhelper.medicationhelper')

def expand_medication_references(input_bundle: Bundle, base_url: str, query_headers: dict = {}) -> Bundle | None:

    returned_resources: list[BundleEntryType] = input_bundle.entry
    output_bundle: dict = deepcopy(input_bundle).dict(exclude_none=True)
    expanded_entries = []

    for entry in returned_resources:
        entry = entry.dict(exclude_none=True) #type: ignore
        resource = entry['resource']
        if 'medicationReference' in resource:
            med_ref = resource['medicationReference']['reference']
            med_lookup = requests.get(f'{base_url}/{med_ref}', headers=query_headers)
            if med_lookup.status_code != 200:
                logger.error(f'The query responded with a status code of {med_lookup.status_code}')
                return None
            med_code_concept = med_lookup.json()['code']
            resource['medicationCodeableConcept'] = med_code_concept
            del resource['medicationReference']
            entry['resource'] = resource
            expanded_entries.append(entry)

    output_bundle['entry'] = expanded_entries

    return Bundle.parse_obj(output_bundle)



from fhirsearchhelper.helpers.documenthelper import expand_document_references
from fhir.resources.R4B.bundle import Bundle


def test_expand_document_references_no_binary() -> None:

    dc_bundle = Bundle.parse_file('./tests/resources/DocumentReferencesNoBinary.json')

    output = expand_document_references(dc_bundle, 'https://hapi.fhir.org/baseR4/')

    assert isinstance(output, Bundle)
    assert output == dc_bundle


'''File to perform filtering of returned FHIR resources using output from gap analysis'''

import json
import logging
from copy import deepcopy
from pathlib import Path

import fhirpathpy
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType

from ..models.models import QuerySearchParams

logger: logging.Logger = logging.getLogger('fhirsearchhelper.fhirfilter')


def filter_bundle(input_bundle: Bundle, search_params: QuerySearchParams, gap_analysis_output: list[str]) -> Bundle:
    '''Function that takes an input bundle, the original search params, and the output from the gap analysis to filter a Bundle'''

    logger.debug('Filtering Bundle using gap analysis output...')

    if not gap_analysis_output:
        return input_bundle

    returned_resources: list[BundleEntryType] = input_bundle.entry
    filtered_entries: list[BundleEntryType] = []
    output_bundle: Bundle = deepcopy(input_bundle)

    for filter_sp in gap_analysis_output:
        filter_sp_value: str | int = search_params.searchParams[filter_sp]
        if '-'  in filter_sp:
            filter_sp = filter_sp[0].lower() + "".join(x.capitalize() for x in filter_sp.lower().split("-"))[1:]
        logger.debug(f'Working on filtering for search parameter {filter_sp}')
        match filter_sp:
            case 'code':
                code_sp_split: list[str] = filter_sp_value.split('%7C')
                if len(code_sp_split) == 2: # Case when there is a | separator
                    logger.debug('Code search parameter value has a |')
                    code_sp_system: str = code_sp_split[0]
                    code_sp_code: str = code_sp_split[1]
                else:
                    logger.debug('Code search parameter contains just the code')
                    code_sp_system = ''
                    code_sp_code = code_sp_split[0]
                for entry in returned_resources:
                    if entry.resource.resource_type == 'MedicationRequest': #type: ignore
                        if code_sp_system and list(filter(lambda x: x.system == code_sp_system and x.code == code_sp_code, entry.resource.medicationCodeableConcept.coding)): # type: ignore
                            logger.debug('Found MedicationRequest that matched both system and code for code element')
                            filtered_entries.append(entry)
                        elif any([coding.code == code_sp_code for coding in entry.resource.medicationCodeableConcept.coding]): # type: ignore
                            logger.debug('Found MedicationRequest that matches code (system was not provided in original query)')
                            filtered_entries.append(entry)
                    else:
                        if code_sp_system and list(filter(lambda x: x.system == code_sp_system and x.code == code_sp_code, entry.resource.code.coding)): # type: ignore
                            logger.debug('Found resource that matched both system and code for code element')
                            filtered_entries.append(entry)
                        elif any([coding.code == code_sp_code for coding in entry.resource.code.coding]): # type: ignore
                            logger.debug('Found resource that matches code (system was not provided in original query)')
                            filtered_entries.append(entry)
            case 'category':
                category_sp_split: list[str] = filter_sp_value.split('|')
                if len(category_sp_split) == 2: # Case when there is a | separator
                    category_sp_system: str = category_sp_split[0]
                    category_sp_code: str = category_sp_split[1]
                else:
                    category_sp_system = ''
                    category_sp_code = category_sp_split[0]
                for entry in returned_resources:
                    if category_sp_system and list(filter(lambda x: x.system == category_sp_system and x.code == category_sp_code, entry.resource.category)): # type: ignore
                        filtered_entries.append(entry)
                    elif any([coding.code == category_sp_code for coding in entry.resource.category.coding]): # type: ignore
                        filtered_entries.append(entry)
            case 'clinicalStatus':
                for entry in returned_resources:
                    if entry.dict(exclude_none=True)['resource'][filter_sp]['coding'][0]['code'] in filter_sp_value.split(','): # type: ignore
                        filtered_entries.append(entry)
            case _:
                for entry in returned_resources:
                    if entry.dict(exclude_none=True)['resource'][filter_sp] == filter_sp_value: # type: ignore
                        filtered_entries.append(entry)

    output_bundle.entry = filtered_entries
    output_bundle.total = len(filtered_entries) # type: ignore

    return output_bundle


def filter_bundle_new(input_bundle: Bundle, search_params: QuerySearchParams, gap_analysis_output: list[str]) -> Bundle:
    '''Function that takes an input bundle, the original search params, and the output from the gap analysis to filter a Bundle'''

    logger.debug('Filtering Bundle using gap analysis output...')

    if not gap_analysis_output:
        return input_bundle

    returned_resources: list[BundleEntryType] = input_bundle.entry
    filtered_entries: list[BundleEntryType] = []
    output_bundle: Bundle = deepcopy(input_bundle)

    with open(f'{Path(__file__).parents[1]}/resources/fhir_r4_search_params.json', 'r') as fin:
        r4_sp_json = json.load(fin)

    for filter_sp in gap_analysis_output:
        filter_sp_value: str | int | dict = search_params.searchParams[filter_sp]
        print('Filter sp value', filter_sp_value)
        filter_sp_fhirpath = r4_sp_json[search_params.resourceType][filter_sp]['fhirpath']

        for entry in returned_resources:
            resource_sp_output = fhirpathpy.evaluate(entry.resource, filter_sp_fhirpath) #type: ignore
            if isinstance(filter_sp_value, str) and '%7C' in filter_sp_value:
                logger.debug('The filter search parameter value is of type code')
                filter_sp_value = {'system': filter_sp_value.split('%7C')[0], 'code': filter_sp_value.split('%7C')[1]}
                logger.debug('Updated filter search parameter value to be a dictionary')
                print(filter_sp_value)
            if filter_sp_value in resource_sp_output:
                filtered_entries.append(entry)

    output_bundle.entry = filtered_entries
    output_bundle.total = len(filtered_entries) # type: ignore

    return output_bundle


'''File to handle all operations around a CapabilityStatement'''

import logging
import os
from pathlib import Path

import requests
from fhir.resources.R4B.capabilitystatement import CapabilityStatement

from ..models.models import SupportedSearchParams

logger: logging.Logger = logging.getLogger('fhirsearchhelper.capabilitystatement')


def load_capability_statement(url: str = None, file_path: str  = None) -> CapabilityStatement: # type: ignore
    '''Function to load a CapabilityStatement into memory'''

    if not url and not file_path:
        raise ValueError('You need to pass a url, an option to specify a preloaded CapabilityStatement, or a file path.')

    if url and file_path:
        logger.info('Defaulting to url...')

    if url:
        try:
            cap_statement: dict = requests.get(url, headers={'Accept': 'application/json'}).json()
        except Exception as exc:
            logger.error('Something went wrong trying to access the CapabilityStatement via URL')
            raise exc

        try:
            cap_statement_object: CapabilityStatement = CapabilityStatement.parse_obj(cap_statement)
        except Exception as exc:
            logger.error('Something went wrong when trying to turn the retrieved cap statement into a CapabilityStatement object')
            raise exc
    else:
        if os.path.isfile(f'{Path(__file__).parents[1]}/capabilitystatements/{file_path}'):
            logger.info(f'Found file {file_path} in the CapabilityStatements folder')
            file_path = f'{Path(__file__).parents[1]}/capabilitystatements/{file_path}'
        cap_statement_object = CapabilityStatement.parse_file(file_path)

    return cap_statement_object


def get_supported_search_params(cs: CapabilityStatement) -> list[SupportedSearchParams]:
    '''Function to pull out supported search parameters from a capability statement'''

    return [SupportedSearchParams(resourceType=resource.type, searchParams=resource.searchParam) for resource in cs.rest[0].resource if resource.searchParam] # type: ignore



'''File to handle all operations around Medication-related Resources'''

import logging
from copy import deepcopy
import json
from .operationoutcomehelper import handle_operation_outcomes

import requests
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.fhirtypes import BundleEntryType

logger: logging.Logger = logging.getLogger('fhirsearchhelper.documenthelper')

def expand_document_references(input_bundle: Bundle, base_url: str, query_headers: dict = {}) -> Bundle | None:

    returned_resources: list[BundleEntryType] = input_bundle.entry
    output_bundle: dict = deepcopy(input_bundle).dict(exclude_none=True)
    expanded_entries = []

    for entry in returned_resources:
        entry = entry.dict(exclude_none=True) #type: ignore
        resource = entry['resource']
        if resource['resourceType'] == 'OperationOutcome':
            handle_operation_outcomes(resource=resource)
            continue
        for i, content in enumerate(resource['content']):
            if 'url' in content['attachment']:
                binary_url = content['attachment']['url']
                binary_url_lookup = requests.get(f'{base_url}/{binary_url}', headers=query_headers)
                if binary_url_lookup.status_code != 200:
                    logger.error(f'The query responded with a status code of {binary_url_lookup.status_code}')
                    if binary_url_lookup.status_code == 403:
                        logger.error('The 403 code typically means your defined scope does not allow for retrieving this resource. Please check your scope to ensure it includes Binary.Read.')
                        if 'WWW-Authenticate' in binary_url_lookup.headers:
                            logger.error(binary_url_lookup.headers['WWW-Authenticate'])
                    return None
                if 'json' in binary_url_lookup.headers['content-type']:
                    content_data = binary_url_lookup.json()['data']
                else:
                    content_data = binary_url_lookup.text
                resource['content'][i]['attachment']['data'] = content_data
                del resource['content'][i]['attachment']['url']
        entry['resource'] = resource
        expanded_entries.append(entry)

    output_bundle['entry'] = expanded_entries

    return Bundle.parse_obj(output_bundle)
